@@ -452,7 +452,139 @@ for filter_name, filter_func in (
 
 
 
-def genconfig(devicename):
+def profiles_merge(source_description, device):
+    """This function recursively merges profiles into the supplied
+    device (or device profile).
+    
+    If there is a "profiles" key in the device, it is read as a list of
+    profiles to merge in, with items in the profiles later in the list
+    taking precedence.  Each of these is merged recursively, allowing
+    them to merge profiles themselves, and/or exclude items (see next).
+
+    If there is a "profiles-exclude" key, it is taken as a structure
+    giving items in the merged set of profiles (above) to be removed
+    before the final merge.
+
+    Once the merge and exclude stages are complete, the supplied device
+    is merged into the merged/excluded profile data (with items in the
+    supplied device taking precedence).
+
+    Keyword arguments:
+
+    source_description -- a textual description of this item (device or
+    device profile), in the form "type: name" (e.g. "device: sw-office"
+    or "device profile: cat3850"), that is used in debugging and error
+    messages.
+
+    device -- a device or device profile definition dictionary that is
+    used as the definition to merge into, as well as read the
+    "profiles" and "profiles-exclude" keys from (if present).
+    """
+
+    # this function makes use of global variables defined outside it so
+    # must appear here in the code
+
+
+    # this dictionary will be used to build up the composite of all the merged
+    # profiles and then exclude items, before merging the device into it at
+    # the end
+
+    profiles_merged = {}
+
+
+    # if there are no profiles to be merged in to this device, return
+
+    if "profiles" not in device:
+        if debug:
+            print("debug: %s complete - no profiles to merge"
+                      % source_description,
+
+                file=sys.stderr)
+
+        return device
+
+
+    # go through the list of profiles to merge
+
+    for profile in device["profiles"]:
+        # check this particular profile is defined and abort with an error
+        # if not
+
+        if profile not in inventory["device-profiles"]:
+            print("error: %s unknown device profile: %s"
+                        % (source_description, profile),
+
+                  file=sys.stderr)
+
+            exit(1)
+
+
+        # get the profile with a recursive call (merging profiles and
+        # excluding items, if the profile does that)
+        #
+        # this will effectively just return the profile itself, if the
+        # profile doesn't do this
+        
+        if debug:
+            print("debug: %s merging device profile: %s"
+                      % (source_description, profile),
+
+                  file=sys.stderr)
+
+        sub_profile_merged = profiles_merge(
+                                    "device profile: %s" % profile,
+                                    inventory["device-profiles"][profile])
+
+
+        # merge a deepcopy() of this profile into the composite profile
+        # we're building of all the merged profiles
+        #
+        # we have to deepcopy() it to avoid changing parts of the merged
+        # structure here and also if we exclude items later
+
+        deepmerge(profiles_merged, deepcopy(sub_profile_merged))
+
+
+    # if we're excluding anything from the imported profiles, we now
+    # remove these from the merged profile we just assembled
+    #
+    # we have to copy the composite merged profile first so we don't
+    # remove anything from the underlying profiles
+    #
+    # we do this inside the "if" statement checking for the "profiles"
+    # key as there is no point excluding anything if we haven't included
+    # any profiles first
+
+    if "profiles-exclude" in device:
+        if debug:
+            print("debug: %s excluding items from merge"
+                        % source_description,
+
+                file=sys.stderr)
+
+        deepremoveitems(profiles_merged, device["profiles-exclude"])
+
+
+    # finally, we merge this device's (or device profile's) definition
+    # dictionary over the top of the merged/excluded profiles; this will give
+    # precedence to the ones in this device's definition
+
+    if debug:
+        print("debug: %s merging" % source_description, file=sys.stderr)
+
+    deepmerge(profiles_merged, device)
+
+
+    # return the device definition with the profiles merged/excluded
+
+    if debug:
+        print("debug: %s complete" % source_description, file=sys.stderr)
+
+    return profiles_merged
+
+
+
+def generate_config(devicename):
     """Generate the configuration for the specified device and write it
     to either standard output or a file, depending on the command line
     options specified.
@@ -479,13 +611,14 @@ def genconfig(devicename):
     if devicename not in inventory["devices"]:
         print("warning: device not found in inventory: %s - skipping"
                   % devicename,
+
               file=sys.stderr)
 
         return False
 
-    device = inventory["devices"][devicename]
+    device_unmerged = inventory["devices"][devicename]
 
-    if device is None:
+    if device_unmerged is None:
         print("warning: device definition empty: %s - skipping" % devicename,
               file=sys.stderr)
 
@@ -496,74 +629,37 @@ def genconfig(devicename):
     # the device definition, so we need to merge them all together and into the
     # explicit device definition itself
 
-    if "profiles" in device:
+    if "profiles" in device_unmerged:
         if "device-profiles" not in inventory:
-            print("error: unknown profile: %s used in device: %s (no "
-                  "profiles defined)" % (device["profile"], devicename),
+            print("error: unknown profile: %s used in device: %s - no "
+                  "profiles defined"
+                      % (device_unmerged["profile"], devicename),
 
                   file=sys.stderr)
 
             exit(1)
 
 
-        # this dictionary will be used to assemble and merge the list of
-        # profiles referenced that we can merge the device configuration into
-
-        profiles_merged = {}
-
-
-        # go through the imported device profiles in order
-
-        for profile in device["profiles"]:
-            # check this particular profile is defined
-
-            if profile not in inventory["device-profiles"]:
-                print("error: unknown profile: %s used in device: %s"
-                          % (profile, devicename),
-
-                      file=sys.stderr)
-
-                exit(1)
-
-
-            # it exists - merge a deep copy of the profile into the merged
-            # profile
-            #
-            # we need to deep copy the profiles as we may remove things from
-            # them due to the 'profiles-exclude' option
-
-            if debug:
-                print("debug: merging device profile: %s" % profile,
-                      file=sys.stderr)
-
-            deepmerge(
-                profiles_merged,
-                deepcopy(inventory["device-profiles"][profile]))
-
-
-        # if we're excluding anything from the imported profiles, we now remove
-        # these from the merged profile we just assembled
-
-        if "profiles-exclude" in inventory["devices"][devicename]:
-            if debug:
-                print("debug: excluding items from device profile",
-                      file=sys.stderr)
-
-            deepremoveitems(profiles_merged, device["profiles-exclude"])
-
-
-        # finally, we merge the device definition dictionary over the top which
-        # will, by default, replace any clashing items, favouring the ones in
-        # the local device definition
-        #
-        # we then replace the device definition with the merged one
+        # call the merge function to recursively merge profiles and exclude
+        # items
 
         if debug:
-            print("debug: merging into device", file=sys.stderr)
+            print("debug: device: %s using profiles" % devicename,
+                      file=sys.stderr)
 
-        deepmerge(profiles_merged, deepcopy(device))
+        device = profiles_merge("device: %s" % devicename, device_unmerged)
 
-        device = profiles_merged
+
+    # we don't have amy profiles for this device, so we just use the device
+    # data as is
+
+    else:
+
+        if debug:
+            print("debug: device: %s not using profiles" % devicename,
+                      file=sys.stderr)
+
+        device = device_unmerged
 
 
     # if the dump device option is enabled, print a YAML version of the device
@@ -698,7 +794,7 @@ for devicename in devicenames:
     if (not quiet) and (len(devicenames) > 1):
         print(devicename)
 
-    complete_success &= genconfig(devicename)
+    complete_success &= generate_config(devicename)
 
 
 exit(0 if complete_success else 1)
